// hlavickove subory *******************************************************
// systemove funkcie
#include <stdio.h>          // header file for standard file i/o
#include <stdlib.h>         // header file for malloc/free

#include "BMPImage.h"

// konstruktor *************************************************************
CBMPImage::CBMPImage()
{

}

// destruktor **************************************************************
CBMPImage::~CBMPImage()
{
 if (bmp_image.data!=NULL) free(bmp_image.data);
}

// nacitanie .bmp suboru
int CBMPImage::BMPImage_load(char *fname)
{
 FILE *file;
 unsigned long size;        // velkost obrazka v bytoch
 unsigned long i;
 unsigned short int planes; // pocet planes v obrazku (potrebne 1)
 unsigned short int bpp;    // pocet bitov na pixel (potrebne 24)
 unsigned char temp;        // pomocna premenna pre bgr-rgb konverziu

 // otvorenie suboru
 if ((file=fopen(fname,"rb"))==NULL) { // 'b' je v linuxe ignorovane
  printf("File not found: %s\n",fname);
  return 0;
 }

 // posun v BMP hlavicke na udaje o sirke a vyske obrazka
 fseek(file,18,SEEK_CUR);

 // nacitanie sirky
 if ((i=fread(&bmp_image.width,4,1,file))!=1) { // 4 byty
  printf("Error reading width from: %s\n",fname);
  return 0;
 }
 printf("sirka obrazka %s: %lu\n",fname,bmp_image.width);

 // nacitanie vysky
 if ((i=fread(&bmp_image.height,4,1,file))!=1) {
  printf("Error reading height from: %s\n",fname);
  return 0;
 }
 printf("vyska obrazka %s: %lu\n",fname,bmp_image.height);

 // vypocet velkosti obrazka (3 byty na pixel)
 size=bmp_image.width*bmp_image.height*3;

 // nacitanie planes
 if ((fread(&planes,2,1,file))!=1) {
  printf("Error reading planes from: %s\n",fname);
  return 0;
 }
 if (planes!=1) {
  printf("Planes from: %s is not 1: %u\n",fname,planes);
  return 0;
 }

 // nacitanie bpp
 if ((i=fread(&bpp,2,1,file))!=1) {
  printf("Error reading bpp from: %s\n",fname);
  return 0;
 }
 if (bpp!=24) {
  printf("Bpp from: %s is not 24: %u\n",fname,bpp);
  return 0;
 }

 // posun na koniec BMP hlavicky
 fseek(file,24,SEEK_CUR);

 // nacitanie samotnych udajov
 bmp_image.data=(unsigned char *) malloc(size);
 if (bmp_image.data==NULL) {
  printf("Error allocating memory for color-corrected image data");
  return 0;
 }
 printf("%lu alokovanych bytov\n",size);

 if ((i=fread(bmp_image.data,size,1,file))!=1) {
  printf("Error reading image data from: %s\n",fname);
  return 0;
 }

 for (i=0;i<size;i+=3) { // korekcia nacitanych farieb (bgr->rgb)
  temp=bmp_image.data[i];
  bmp_image.data[i]=bmp_image.data[i+2];
  bmp_image.data[i+2]=temp;
 }

 return 1;
}

unsigned long CBMPImage::get_width()
{
 return bmp_image.width;
}

unsigned long CBMPImage::get_height()
{
 return bmp_image.height;
}

unsigned char *CBMPImage::get_data()
{
 return bmp_image.data;
}

