// hlavickove subory *******************************************************
// systemove funkcie
#include <math.h>

// GL a GLU kniznica
#include <GL/gl.h>
#include <GL/glu.h>

#include "glFrustum.h"

// zoznam stran Frustum
enum FrustumSide
{
 RIGHT =0, // prava strana Frustum
 LEFT  =1, // lava strana Frustum
 BOTTOM=2, // spodna strana Frustum
 TOP   =3, // horna strana Frustum
 BACK  =4, // zadna strana Frustum
 FRONT =5  // predna strana Frustum
}; 

// zoznam hodnot strany Frustum
enum PlaneData
{
 A=0, // X hodnota normaly
 B=1, // Y hodnota normaly
 C=2, // Z hodnota normaly
 D=3  // vzdialenost strany od pociatku
};

// normalizacia strany Frustum
void NormalizePlane(GLfloat frustum[6][4],GLuint side)
{
 // vypocet magnitudy normaly strany
 // magnitude = sqrt( x^2 + y^2 + z^2)
 GLfloat magnitude=(GLfloat) sqrt(frustum[side][A]*frustum[side][A]+
                                  frustum[side][B]*frustum[side][B]+
                                  frustum[side][C]*frustum[side][C]);

 // delenie hodnot strany magnitudou jej normaly
 frustum[side][A]/=magnitude;
 frustum[side][B]/=magnitude;
 frustum[side][C]/=magnitude;
 frustum[side][D]/=magnitude;
}

// funkcia pre vypocet Frustum
void FrustumCulling::FrustumUpdate()
{
 GLfloat proj[16]; // nacitanie projekcnej matice
 GLfloat modl[16]; // nacitanie modelview matice
 GLfloat clip[16]; // vypocet orezavecich rovin

 // ziskanie projekcnej matice
 glGetFloatv(GL_PROJECTION_MATRIX,proj);

 // nacitanie modelview matice
 glGetFloatv(GL_MODELVIEW_MATRIX,modl);

 // ziskanie orezavacich rovin kombinaciou matic
 clip[ 0]=modl[ 0]*proj[ 0]+modl[ 1]*proj[ 4]+modl[ 2]*proj[ 8]+modl[ 3]*proj[12];
 clip[ 1]=modl[ 0]*proj[ 1]+modl[ 1]*proj[ 5]+modl[ 2]*proj[ 9]+modl[ 3]*proj[13];
 clip[ 2]=modl[ 0]*proj[ 2]+modl[ 1]*proj[ 6]+modl[ 2]*proj[10]+modl[ 3]*proj[14];
 clip[ 3]=modl[ 0]*proj[ 3]+modl[ 1]*proj[ 7]+modl[ 2]*proj[11]+modl[ 3]*proj[15];

 clip[ 4]=modl[ 4]*proj[ 0]+modl[ 5]*proj[ 4]+modl[ 6]*proj[ 8]+modl[ 7]*proj[12];
 clip[ 5]=modl[ 4]*proj[ 1]+modl[ 5]*proj[ 5]+modl[ 6]*proj[ 9]+modl[ 7]*proj[13];
 clip[ 6]=modl[ 4]*proj[ 2]+modl[ 5]*proj[ 6]+modl[ 6]*proj[10]+modl[ 7]*proj[14];
 clip[ 7]=modl[ 4]*proj[ 3]+modl[ 5]*proj[ 7]+modl[ 6]*proj[11]+modl[ 7]*proj[15];

 clip[ 8]=modl[ 8]*proj[ 0]+modl[ 9]*proj[ 4]+modl[10]*proj[ 8]+modl[11]*proj[12];
 clip[ 9]=modl[ 8]*proj[ 1]+modl[ 9]*proj[ 5]+modl[10]*proj[ 9]+modl[11]*proj[13];
 clip[10]=modl[ 8]*proj[ 2]+modl[ 9]*proj[ 6]+modl[10]*proj[10]+modl[11]*proj[14];
 clip[11]=modl[ 8]*proj[ 3]+modl[ 9]*proj[ 7]+modl[10]*proj[11]+modl[11]*proj[15];

 clip[12]=modl[12]*proj[ 0]+modl[13]*proj[ 4]+modl[14]*proj[ 8]+modl[15]*proj[12];
 clip[13]=modl[12]*proj[ 1]+modl[13]*proj[ 5]+modl[14]*proj[ 9]+modl[15]*proj[13];
 clip[14]=modl[12]*proj[ 2]+modl[13]*proj[ 6]+modl[14]*proj[10]+modl[15]*proj[14];
 clip[15]=modl[12]*proj[ 3]+modl[13]*proj[ 7]+modl[14]*proj[11]+modl[15]*proj[15];

 // ziskanie stran orezavacich rovin
 m_Frustum[RIGHT][A]=clip[ 3]-clip[ 0];
 m_Frustum[RIGHT][B]=clip[ 7]-clip[ 4];
 m_Frustum[RIGHT][C]=clip[11]-clip[ 8];
 m_Frustum[RIGHT][D]=clip[15]-clip[12];
 // normalizujeme stranu zo ziskanej normaly (A,B,C) a vzdialenosti (D)
 NormalizePlane(m_Frustum,RIGHT);

 m_Frustum[LEFT][A]=clip[ 3]+clip[ 0];
 m_Frustum[LEFT][B]=clip[ 7]+clip[ 4];
 m_Frustum[LEFT][C]=clip[11]+clip[ 8];
 m_Frustum[LEFT][D]=clip[15]+clip[12];
 NormalizePlane(m_Frustum,LEFT);

 m_Frustum[BOTTOM][A]=clip[ 3]+clip[ 1];
 m_Frustum[BOTTOM][B]=clip[ 7]+clip[ 5];
 m_Frustum[BOTTOM][C]=clip[11]+clip[ 9];
 m_Frustum[BOTTOM][D]=clip[15]+clip[13];
 NormalizePlane(m_Frustum,BOTTOM);

 m_Frustum[TOP][A]=clip[ 3]-clip[ 1];
 m_Frustum[TOP][B]=clip[ 7]-clip[ 5];
 m_Frustum[TOP][C]=clip[11]-clip[ 9];
 m_Frustum[TOP][D]=clip[15]-clip[13];
 NormalizePlane(m_Frustum,TOP);

 m_Frustum[BACK][A]=clip[ 3]-clip[ 2];
 m_Frustum[BACK][B]=clip[ 7]-clip[ 6];
 m_Frustum[BACK][C]=clip[11]-clip[10];
 m_Frustum[BACK][D]=clip[15]-clip[14];
 NormalizePlane(m_Frustum,BACK);

 m_Frustum[FRONT][A]=clip[ 3]+clip[ 2];
 m_Frustum[FRONT][B]=clip[ 7]+clip[ 6];
 m_Frustum[FRONT][C]=clip[11]+clip[10];
 m_Frustum[FRONT][D]=clip[15]+clip[14];
 NormalizePlane(m_Frustum,FRONT);
}

/*
  pre test, ci je bod, gula, kocka... vnutri Frustum pouzivame predpoklad,
  ze vsetky normaly stran Frustum smeruju dovnutra Frustum priestoru a teda
  pokial je bod pred vsetkymi 6-timi stranami, je vnutri Frustum
*/

// test, ci je bod vo Frustum
bool FrustumCulling::PointInFrustum(GLfloat x,GLfloat y,GLfloat z)
{
 // test vsetkych 6-tich stran
 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*x+m_Frustum[i][B]*y+m_Frustum[i][C]*z+m_Frustum[i][D]<=-10) {
   // bod je za stranou => je vonku z Frustum
   return false;
  }
 }
 // bod je pred vsetkymi 6-timi stranami => je vo Frustum
 return true;
}

// test, ci je gula vo Frustum
bool FrustumCulling::SphereInFrustum(GLfloat x,GLfloat y,GLfloat z,GLfloat radius)
{
 // test vsetkych 6-tich stran
 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*x+m_Frustum[i][B]*y+m_Frustum[i][C]*z+m_Frustum[i][D]<=-radius) {
   // stred gule+jej polomer je za stranou => je vonku z Frustum
   return false;
  }
 }
 // gula je pred vsetkymi 6-timi stranami => je vo Frustum
 return true;
}

// test, ci je kocka vo Frustum
bool FrustumCulling::CubeInFrustum(GLfloat x,GLfloat y,GLfloat z,GLfloat size)
{
 // Basically, what is going on is, that we are given the center of the cube,
 // and half the length.  Think of it like a radius.  Then we checking each point
 // in the cube and seeing if it is inside the frustum.  If a point is found in front
 // of a side, then we skip to the next side.  If we get to a plane that does NOT have
 // a point in front of it, then it will return false.

 // This will sometimes say that a cube is inside the frustum when it isn't.
 // This happens when all the corners of the bounding box are not behind any one plane.
 // This is rare and shouldn't effect the overall rendering speed.

 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  // stred kocky+1/2 jej velkosti je za stranou => je vonku z Frustum
  return false;
 }
 // kocka je pred vsetkymi 6-timi stranami => je vo Frustum
 return true;
}

// test, ci je kocka vo Frustum
GLuint FrustumCulling::CubeInFrustum2(GLfloat x,GLfloat y,GLfloat z,GLfloat size)
{
 int je_dnu=2; // je_dnu=2-kocka je cela dnu
 int bol=0;
 for (i=0;i<6;i++) {
  bol=0;
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   { bol=1; if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   { bol=1; if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   { bol=1; if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   { bol=1; if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   { bol=1; if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   { bol=1; if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   { bol=1; if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   {        if (je_dnu==1) continue; }
   else { je_dnu=1; if (bol) continue; }

  if (je_dnu==2) continue;

  // stred kocky+1/2 jej velkosti je za stranou => je vonku z Frustum
  return 0;
 }
 // kocka je pred vsetkymi 6-timi stranami => je vo Frustum
 return je_dnu;
}

void FrustumCulling::CullTriangles(GLfloat *vertex,GLfloat *vertex1,
                                   GLfloat *normal,GLfloat *normal1,
                                   GLuint all_tris_num,GLuint *cull_tris_num)
{
 b=0;

 for (a=0;a<all_tris_num*9;a+=9) {

  if (PointInFrustum(*(vertex+a),  *(vertex+a+1),*(vertex+a+2))||
      PointInFrustum(*(vertex+a+3),*(vertex+a+4),*(vertex+a+5))||
      PointInFrustum(*(vertex+a+6),*(vertex+a+7),*(vertex+a+8))) {
   *(vertex1+b  )=*(vertex+a);   *(vertex1+b+1)=*(vertex+a+1); *(vertex1+b+2)=*(vertex+a+2);
   *(vertex1+b+3)=*(vertex+a+3); *(vertex1+b+4)=*(vertex+a+4); *(vertex1+b+5)=*(vertex+a+5);
   *(vertex1+b+6)=*(vertex+a+6); *(vertex1+b+7)=*(vertex+a+7); *(vertex1+b+8)=*(vertex+a+8);
   *(normal1+b  )=*(normal+a);   *(normal1+b+1)=*(normal+a+1); *(normal1+b+2)=*(normal+a+2);
   *(normal1+b+3)=*(normal+a+3); *(normal1+b+4)=*(normal+a+4); *(normal1+b+5)=*(normal+a+5);
   *(normal1+b+6)=*(normal+a+6); *(normal1+b+7)=*(normal+a+7); *(normal1+b+8)=*(normal+a+8);
   b+=9;
  }

 }

 *cull_tris_num=b/9;
}

/*
  tento kod vznikol po uprave zverejneneho kodu autora:

  Ben Humphrey (DigiBen)
  Game Programmer
  DigiBen@GameTutorials.com
  Co-Web Host of www.GameTutorials.com
  date: 11/26/01
*/

