// GL a GLU kniznica
#include <GL/gl.h>
#include <GL/glu.h>

#include <stdio.h>
#include <malloc.h>

#include "glTerrain.h"

// konstruktor *************************************************************
glTerrain::glTerrain()
{
 all_triangles=0; cull_triangles=all_triangles;

 vertex=NULL; vertex1=NULL;
 normal=NULL; normal1=NULL;
}

// destruktor **************************************************************
glTerrain::~glTerrain()
{
 if (vertex!=NULL) free(vertex);
 if (vertex1!=NULL) free(vertex1);

 if (normal!=NULL) free(normal);
 if (normal1!=NULL) free(normal1);
}

void NormaliseNormal(GLfloat n1,GLfloat n2,GLfloat n3,
                     GLfloat *a,GLfloat *b,GLfloat *c)
{
 GLfloat d;
 d=n1; if (d<n2) d=n2; if (d<n3) d=n3;
 *a=n1/d; *b=n2/d; *c=n3/d;
}

void CalcNormal(GLfloat p1,GLfloat p2,GLfloat p3,
                GLfloat q1,GLfloat q2,GLfloat q3,
                GLfloat r1,GLfloat r2,GLfloat r3,
                GLfloat *a,GLfloat *b,GLfloat *c)
{
 GLfloat v11,v12,v13;
 GLfloat v21,v22,v23;
 GLfloat n1,n2,n3,nn1,nn2,nn3;

 v11=q1-p1; v12=q2-p2; v13=q3-p3;
 v21=r1-p1; v22=r2-p2; v23=r3-p3;

 n1=v12*v23-v13*v22;
 n2=v13*v21-v11*v23;
 n3=v11*v22-v12*v21;

 NormaliseNormal(n1,n2,n3,&nn1,&nn2,&nn3);

 *a=nn1; *b=nn2; *c=nn3;
}

// priprava terenu *********************************************************
int glTerrain::terrain_create(unsigned char *bmp_data)
{
 ta=0.0f; tb=0.0f;
 t1=0; t2=0; t3=0; t4=0;
 q=1024.0f; w=8.0f; e=q/w; r=3*1024;

 vertex=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex) {
  fprintf(stderr,"Error allocating memory for vertexes!\n");
  return(1);
 }
 vertex1=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex1) {
  fprintf(stderr,"Error allocating memory for vertexes1!\n");
  return(1);
 }

 normal=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex) {
  fprintf(stderr,"Error allocating memory for normals!\n");
  return(1);
 }
 normal1=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex1) {
  fprintf(stderr,"Error allocating memory for normals1!\n");
  return(1);
 }

 a=0;

 for (ty=0.0f;ty<(q-w);ty+=w) {
  for (tx=0.0f;tx<(q-w);tx+=w) {
   t1=(GLuint) (3*tx+(r*ty));
   t2=(GLuint) (3*(tx+w)+(r*ty));
   t3=(GLuint) (3*(tx+w)+(r*(ty+w)));
   t4=(GLuint) (3*tx+(r*(ty+w)));

   CalcNormal(2.0f*(tx),  (GLfloat) bmp_data[t1],2.0f*(-ty),
              2.0f*(tx+w),(GLfloat) bmp_data[t2],2.0f*(-ty),
              2.0f*(tx),  (GLfloat) bmp_data[t4],2.0f*(-ty-w),
              &n1,&n2,&n3);

   //glTexCoord2f(ta,tb);
   //glVertex3f(tx  ,(GLfloat) bmp_data[t1],-ty);
   *(vertex+a)=tx; *(vertex+a+1)=(GLfloat) bmp_data[t1]; *(vertex+a+2)=-ty;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;

   //glTexCoord2f(ta+1.0f/e,tb);
   //gl*(vertex+3f(tx+w,(GLfloat) bmp_data[t2],-ty);
   *(vertex+a)=tx+w; *(vertex+a+1)=(GLfloat) bmp_data[t2]; *(vertex+a+2)=-ty;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;

   //glTexCoord2f(ta,tb+1.0f/e);
   //gl*(vertex+3f(tx  ,(GLfloat) bmp_data[t4],-ty-w);
   *(vertex+a)=tx; *(vertex+a+1)=(GLfloat) bmp_data[t4]; *(vertex+a+2)=-ty-w;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;

   CalcNormal(2.0f*(tx+w),(GLfloat) bmp_data[t2],2.0f*(-ty),
              2.0f*(tx+w),(GLfloat) bmp_data[t3],2.0f*(-ty-w),
              2.0f*(tx),  (GLfloat) bmp_data[t4],2.0f*(-ty-w),
              &n1,&n2,&n3);

   //glTexCoord2f(ta+1.0f/e,tb);
   //gl*(vertex+3f(tx+w,(GLfloat) bmp_data[t2],-ty);
   *(vertex+a)=tx+w; *(vertex+a+1)=(GLfloat) bmp_data[t2]; *(vertex+a+2)=-ty;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;

   //glTexCoord2f(ta+1.0f/e,tb+1.0f/e);
   //gl*(vertex+3f(tx+w,(GLfloat) bmp_data[t3],-ty-w);
   *(vertex+a)=tx+w; *(vertex+a+1)=(GLfloat) bmp_data[t3]; *(vertex+a+2)=-ty-w;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;

   //glTexCoord2f(ta,tb+1.0f/e);
   //gl*(vertex+3f(tx  ,(GLfloat) bmp_data[t4],-ty-w);
   *(vertex+a)=tx; *(vertex+a+1)=(GLfloat) bmp_data[t4]; *(vertex+a+2)=-ty-w;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;

   ta+=1.0f/e;
  }
  ta=0.0f; tb+=1.0f/e;
 }

 all_triangles=a/9; cull_triangles=all_triangles;

 for (a=0;a<all_triangles*9;a+=9) {
  *(vertex+a)*=2.0f;   *(vertex+a+2)*=2.0f;
  *(vertex+a+3)*=2.0f; *(vertex+a+5)*=2.0f;
  *(vertex+a+6)*=2.0f; *(vertex+a+8)*=2.0f;

  *(vertex1+a)  =*(vertex+a); *(vertex1+a+1)=*(vertex+a+1); *(vertex1+a+2)=*(vertex+a+2);
  *(vertex1+a+3)=*(vertex+a+3); *(vertex1+a+4)=*(vertex+a+4); *(vertex1+a+5)=*(vertex+a+5);
  *(vertex1+a+6)=*(vertex+a+6); *(vertex1+a+7)=*(vertex+a+7); *(vertex1+a+8)=*(vertex+a+8);
 }

 glVertexPointer(3,GL_FLOAT,0,vertex1);
 glEnableClientState(GL_VERTEX_ARRAY);
 glNormalPointer(GL_FLOAT,0,normal1);
 glEnableClientState(GL_NORMAL_ARRAY);

 return(0);
}

// vratenie pola trojuholnikov *********************************************
GLfloat *glTerrain::terrain_getTriangles(void)
{
 return vertex;
}

// vratenie pola trojuholnikov *********************************************
GLfloat *glTerrain::terrain_getTriangles1(void)
{
 return vertex1;
}

// vratenie pola normal ****************************************************
GLfloat *glTerrain::terrain_getNormals(void)
{
 return normal;
}

// vratenie pola normal ****************************************************
GLfloat *glTerrain::terrain_getNormals1(void)
{
 return normal1;
}

// kreslenie terenu ********************************************************
void glTerrain::terrain_render(GLuint tris_num)
{
 glDrawArrays(GL_TRIANGLES,0,3*tris_num);
}

// kreslenie normal ********************************************************
void glTerrain::normals_render(GLuint tris_num)
{
 glDisable(GL_LIGHTING);
 glColor4f(1.0f,0.0f,0.0f,1.0f);
 glBegin(GL_LINES);
  for (a=0;a<tris_num*9;a+=9) {
   glVertex3fv(vertex1+a);
   glVertex3f(*(vertex1+a)+*(normal1+a)*20.0f,
              *(vertex1+a+1)+*(normal1+a+1)*20.0f,
              *(vertex1+a+2)+*(normal1+a+2)*20.0f);
  }
 glEnd();
 glEnable(GL_LIGHTING);
}

